/*  VQF XMMS Plugin 
 *  Copyright (C) 2000  Mel Gorman (mel@skynet.ie) 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "vqf.h"
#include <string.h>
#include <pthread.h>

/* Handle to file currently playing */
VQFFile vqfCurrent;

/* Decode thread */
static pthread_t decodeThread;

#define getSongError() \
{ \
    TRACE("getSongInfo: In error handler\n"); \
    strcpy (fileInfo->header.Name, "ERROR"); \
    fileInfo->length = 0; \
    TRACE("getSongInfo: closing file\n"); \
    if (fileInfo->file && closeFile) bclose(fileInfo->file); \
    TRACE("getSongInfo: leaving error handler\n"); \
    return -1; \
} 
                                                                                                
extern "C" {
/* XMMS Struct */
InputPlugin vqfIP =
{
  NULL,
  NULL,
  "VQF Player 0.9",
  vqfInit,
  vqfAbout,
  NULL,
  vqfFile,
  NULL,
  vqfPlay,
  vqfStop,
  vqfPause,
  vqfSeek,
  NULL,
  vqfGetTime,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  vqfGetInfo,
  vqfInfoBox,
  NULL
};

/* Function to return a handle to the plugin */
InputPlugin *get_iplugin_info(void) { return &vqfIP; }
}

/* Gets text data from header */ 
static int getTextInfo(char *buf, char *ID, CHeaderManager *hManage)
{
  try
  {
    CStringChunk sChunk = hManage->GetPrimaryChunk(std::string(ID));
    strcpy(buf, sChunk.GetString().c_str());
  } catch (CHeaderManager::err_FailGetChunk)
  {
    sprintf(buf, "");
    return 0;
  }

  return 1;
}

/* DSiz is related to the length of the song in bytes */
void GetDSiz(CHeaderManager *hManage, int *DSiz, char *filename)
{
  try {
    CDsizChunk dChunk(hManage->GetPrimaryChunk("DSIZ"));
    *DSiz = dChunk.GetInt();
  } catch (CHeaderManager::err_FailGetChunk)
  {
    TRACE("vqfPlay: Failed to get dsiz\n");
    struct stat buf;
    if (stat(filename, &buf) != -1) *DSiz = buf.st_size - 128;
    else *DSiz = 0;
  }
}                        

/* Inilialise plugin */
static void vqfInit(void) 
{ 
  /* strcpy(mel->stomach, offLicense->get("Beer")); */ 
  TRACE("vqfInit: Initialising\n");
}

/* Display about box */
static void vqfAbout(void)
{
  static GtkWidget *messageBox = NULL;
  if (!messageBox)
  {
    messageBox = xmms_show_message("About", "VQF Player 0.9\n\n(c) Mel Gorman 2000", "Ok", FALSE, NULL, NULL);
    gtk_signal_connect(GTK_OBJECT(messageBox), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &messageBox);
  }
}


/* Check if this file belongs to us
 * Easist is to check extension. 
 * Alternatively, check if the first few bytes of a file is TWINVQ */
static int  vqfFile(char *filename)
{
  char *ext = strrchr(filename, '.');
  if (ext && !strcasecmp(ext, ".vqf")) return TRUE;

  return FALSE;
}

/* Taken from the TwinVQ SDK directly */
/* Not quiet sure what it does, but I would guess it converts
 * from RAW to PCM. If I'm wrong, I'm not pushed. But if someone
 * could explain it to me, I'd appreciate it :-) */
void frtobuf(float out[],     
             short  bufout[],
             int    frameSize,
             int    numChannels)
{
   /*--- Variables ---*/
   int ismp, ich, uflag, lflag;
   float *ptr;
   register float dtmp;

   uflag = lflag = 0;

   for ( ich=0; ich<numChannels; ich++ )
   {
     ptr = out+ich*frameSize;
     for ( ismp=0; ismp<frameSize; ismp++ )
     {
       dtmp = ptr[ismp];
       if ( dtmp >= 0. )
       {
         if ( dtmp > 32700. )
         {
           dtmp = 32700.;
           uflag = 1;
         }
         
         bufout[ismp*numChannels+ich] = (short)(dtmp+0.5);
       }
       else
       {
         if ( dtmp < -32700. )
         {
           dtmp = -32700.;
           lflag = 1;
         }
         bufout[ismp*numChannels+ich] = (short)(dtmp-0.5);
       }
     }
   }
}

/* Once decoding starts, this function begins looping
 * and going round and round and round and DOES IT EVER END? 
 * Send beer to Mel, save him from insanity! */
static void *vqfPlayLoop(void *arg) 
{
  TRACE("vqfPlayLoop: Entering\n");
  bool lastFrame = false;
  int bytesPerFrame = vqfCurrent.frameSize * vqfCurrent.numChannels;
  
  TRACE("vqfPlayLoop: Allocing memory\n");
  /* Output buffer for the decode */
  float *decodeOut = new float[ bytesPerFrame ];
  if (decodeOut == NULL)
  {
    printf("vqfPlayLoop: Failed to alloc\n");
    pthread_exit(NULL);
  }

  /* Output buffer for audio */
  short *audioOut =  new short[ bytesPerFrame ]; 
  if (audioOut == NULL)
  {
    printf("vqfPlayLoop: Failed to alloc\n");
    delete decodeOut;
    pthread_exit(NULL);
  }

  TRACE("vqfPlayLoop: Init BSReader\n");
  if (TvqInitBsReader(&(vqfCurrent.header)))
  {
    printf("vqfPlayLoop: Failed to init Bs reader\n");
    pthread_exit(NULL);
  }

  /* Check we haven't been told to stop early */
  int wFlag=2;
  int seekFrame=0;
  while(vqfCurrent.playing == true && !lastFrame)
  {
    /* Check to see should we seek */
    if (vqfCurrent.seekTime != -1)
    {
      TRACE("vqfPlayLoop: Seeking\n");
      seekFrame = int(double(vqfCurrent.seekTime - (vqfIP.output->written_time() / 1000)) / vqfCurrent.frameSize * vqfCurrent.samplesPerSec);
      vqfIP.output->flush(vqfCurrent.seekTime * 1000);
      TvqSkipFrame(vqfCurrent.file, seekFrame);
      vqfCurrent.seekTime = -1;
    }

    /* Read in a frame */
    TRACE("vqfPlayLoop: Reading BSFrame\n");
    if (TvqReadBsFrame(&(vqfCurrent.index), vqfCurrent.file) == 0)
      lastFrame = true;

    /* Decode the current frame */
    TRACE("vqfPlayLoop: Decoding current frame\n");
    TvqDecodeFrame(&(vqfCurrent.index), decodeOut);
 
    /* Convert to PCM */
    TRACE("vqfPlayLoop: Converting to PCM\n");
    frtobuf(decodeOut, audioOut, vqfCurrent.frameSize, vqfCurrent.numChannels);
    
    /* Pause if necessary */
    while (vqfIP.output->buffer_free() < bytesPerFrame*2 &&
           vqfCurrent.playing == true &&
           vqfCurrent.seekTime == -1) xmms_usleep(10000);

    /* Send to audio plugin */
    vqfCurrent.currentFrame++;
    TRACE("vqfPlayLoop: Sending to Audio\n");

    if (wFlag == 0)
    {
      vqfIP.add_vis_pcm(vqfIP.output->written_time(),
                         (8 * sizeof(short) == 16) ? FMT_S16_LE: FMT_U8,
                         vqfCurrent.numChannels,
                         bytesPerFrame*sizeof(short),
                         audioOut);
      vqfIP.output->write_audio(audioOut, bytesPerFrame*sizeof(short));
    }
    wFlag = wFlag -1 > 0 ? wFlag - 1 : 0;

  }

  /* Decoding is complete */
  TRACE("vqfPlayLoop: Decoding complete\n");
  TvqTerminate(&(vqfCurrent.index));
  bclose(vqfCurrent.file);
  vqfCurrent.playing=false;
  
  /* Give back memory */
  delete decodeOut;
  delete audioOut;

  pthread_exit(NULL);
}

/* Play a song */
static void vqfPlay(char *filename) 
{ 
  /* Happy days. A file is to be played 
   * Send Mel a Six pack in gratidute! */

  struct stat buf;
  stat(filename, &buf);

  /* Get song information */
  if (getSongInfo(filename, &vqfCurrent, false) == -1) return;
  vqfCurrent.seekTime = -1;

  /* Initialise TwinVQ Dll */
  TRACE("vqfPlay: Initializing dll\n");
  if (TvqInitialize(&(vqfCurrent.header), &(vqfCurrent.index), 1) )
  {
    printf("VQF: Failed to initilaze\n");
    return;
  }
  
  /* Get more information. Send Mel more Beer */
  TRACE("vqfPlay: Getting additional information\n");
  vqfCurrent.bitsPerSample =    TvqGetBitRate();
  vqfCurrent.samplesPerSec =    (int)TvqGetSamplingRate();
  vqfCurrent.numChannels =      TvqGetNumChannels();
  vqfCurrent.frameSize =        TvqGetFrameSize();
  vqfCurrent.totalFrames =      TvqGetNumFrames();
  vqfCurrent.currentFrame =     0;
  
  /* Open the output plugin */
  TRACE("vqfPlay: Opening Audio\n");
  if (vqfIP.output->open_audio(((8 * sizeof(short)) == 16 ? FMT_S16_LE : FMT_U8),
                             vqfCurrent.samplesPerSec,
                             vqfCurrent.numChannels) == 0)
  {
    return;
  }

  /* Set up vis information */
  vqfIP.set_info(vqfCurrent.header.Name,                                    /* File name */
                 vqfCurrent.length,    					    /* Length */
                 vqfCurrent.samplesPerSec * vqfCurrent.numChannels * (vqfCurrent.bitsPerSample / 8), /* Rate */
                 vqfCurrent.samplesPerSec,                                  /* Samples per sec */
                 vqfCurrent.numChannels);                                   /* Channels */

  /* Enter Decode loop */
  TRACE("vqfPlay: Creating decode thread\n");
  vqfCurrent.playing = true;

  pthread_create(&decodeThread, NULL, vqfPlayLoop, NULL);
  return;

}

/* Stop output */
static void vqfStop(void) 
{ 
  TRACE("vqfStop: Stop called\n");
  
  /* BANG, you're dead */
  vqfCurrent.playing=false;

  /* Rejoin thread if necessary */
  pthread_join(decodeThread, NULL);
  
  /* Close the audio */
  vqfIP.output->close_audio();

}

/* Pause output */
static void vqfPause(short p) 
{ 
  TRACE("vqfPause: Pause called\n");
  vqfIP.output->pause(p); 
}

/* Seek in VQF */
static void vqfSeek(int time) 
{ 
  vqfCurrent.seekTime = time; 
  while (vqfCurrent.seekTime != -1) xmms_usleep(10000);
} 

/* Get how much has been played so far */
static int  vqfGetTime(void) 
{ 
  TRACE("vqfGetTime: Returning time\n");
  if (vqfCurrent.playing == true) return vqfIP.output->output_time();
  else return -1;
}

static void vqfGetInfo(char *filename, char **title, int *length) 
{
  VQFFile fileInfo;
  TRACE("vqfGetInfo: Calling getSongInfo()\n");
  getSongInfo(filename, &fileInfo);

  /* Copy it in */
  TRACE("vqfGetInfo: Copying name\n");
  *title = (gchar *)g_malloc(strlen(fileInfo.header.Name)+1);
  strcpy(*title, fileInfo.header.Name);
  *length = fileInfo.length;
}

int getSongInfo(char *filename, VQFFile *fileInfo, bool closeFile)
{ 

  /* Returns information from the song */
  gchar *name;
  INDEX index;
  CChunkChunk *tChunk;
  CHeaderManager *hManage;

  /* Open FILE */
  TRACE("getSongInfo: Opening file\n");
  fileInfo->file = bopen(filename, "r");
  if (fileInfo->file == NULL) getSongError();
  
  /* Read in header */
  TRACE("vqfGetInfo: Opening BS header\n");
  tChunk = TvqGetBsHeaderInfo(fileInfo->file);
  if (tChunk == NULL) getSongError();

  /* Parse out header */
  TRACE("vqfGetInfo: Getting header manager\n");
  hManage = CHeaderManager::Create(*tChunk);
  if (hManage == NULL) getSongError();
  delete tChunk;

  /* Get the title from the header */
  TRACE("vqfGetInfo: Getting Primary Chunk\n");
  sprintf(fileInfo->header.ID, "%s", hManage->GetID().c_str());
  CCommChunk commChunk(hManage->GetPrimaryChunk("COMM"));
  fileInfo->header.channelMode   = commChunk.GetChannelMode();
  fileInfo->header.bitRate       = commChunk.GetBitRate();
  fileInfo->header.samplingRate  = commChunk.GetSamplingRate();
  fileInfo->header.securityLevel = commChunk.GetSecurityLevel();

  if (getTextInfo(fileInfo->header.Name, "NAME", hManage) == 0)
    strcpy(fileInfo->header.Name, strrchr(filename, '/')+1);
  getTextInfo(fileInfo->header.Auth, "AUTH", hManage);
  getTextInfo(fileInfo->header.Comt, "COMT", hManage);
  getTextInfo(fileInfo->header.Cpyr, "(c) ", hManage);
  getTextInfo(fileInfo->header.File, "FILE", hManage);

  TRACE("vqfGetInfo: Getting dsiz\n");
  GetDSiz(hManage, &(fileInfo->header.Dsiz), filename);

  /* Calculate length */
  TRACE("vqfGetInfo: Calcaulating length\n");
  fileInfo->length = fileInfo->header.Dsiz / (fileInfo->header.bitRate / 8) - 3000; 

  /* Close and return */
  TRACE("vqfGetInfo: Returning\n");
  delete hManage;
  if (closeFile) bclose(fileInfo->file);

  return 1;
}

